import torch

try:
    import triton
    import triton.language as tl
    _TRITON_AVAILABLE = True

except Exception:
    # Triton not available on this system – fall back to PyTorch ops.
    # If the mask has many unique values, attention will be slow.
    _TRITON_AVAILABLE = False
_TRITON_AVAILABLE = False


def _multi_linear_attn_no_mask(query, key, value, dropout_p, kv_cache):
    key_value_store = key.transpose(-2, -1) @ value + kv_cache  # [..., d_k, d_v]
    key_value_store_view = torch.dropout(key_value_store, dropout_p, train=True)
    out = query @ key_value_store_view
    return out, key_value_store


def _multi_linear_attn_mask_torch(query, key, value, mask, dropout_p, kv_cache):
    key = key.unsqueeze(-1)  # [..., S, d_k, 1]
    value = value.unsqueeze(-2)  # [..., S, 1, d_v]
    key_value_store = key @ value  # [..., S, d_k, d_v]
    key_value_store = key_value_store.cumsum(dim=-3)[..., mask, :, :] + kv_cache.unsqueeze(-3)
    key_value_store = torch.dropout(key_value_store, dropout_p, train=True)
    out = (query.unsqueeze(-2) @ key_value_store).squeeze(-2)
    key_value_store = key_value_store[..., -1, :, :]
    return out, key_value_store


def pad_to_pow2(x: torch.Tensor, dims=None) -> torch.Tensor:
    shape = torch.as_tensor(x.shape, device=x.device, dtype=torch.long)

    if dims is None:
        mask = torch.ones_like(shape, dtype=torch.bool)
    else:
        idx = torch.remainder(
            torch.as_tensor(dims, device=x.device, dtype=torch.long), x.dim()
        )
        mask = torch.zeros_like(shape, dtype=torch.bool)
        mask[idx] = True

    next_pow2 = (2 ** torch.ceil(torch.log2(shape.float()))).long()
    target = torch.where(mask, next_pow2, shape)

    out = x.new_zeros(tuple(target.tolist()))
    out[tuple(slice(0, int(s)) for s in shape)] = x
    return out

if _TRITON_AVAILABLE:
    def _multi_linear_attn_mask_triton(query, key, value, mask, dropout_p, kv_cache):
        B, H, S, DK = query.shape
        R, DV          = value.shape[-2:]

        q = query.contiguous()
        k = key.contiguous()
        v = value.contiguous()
        mask = mask.contiguous()
        kv_cache = kv_cache.contiguous()

        out = torch.empty((B, H, S, DV), device=query.device, dtype=query.dtype)
        kv_out = torch.empty((B, H, DK, DV), device=query.device, dtype=query.dtype)

        grid = (B * H,)

        base_seed = torch.randint(2**31, (1,), device=query.device, dtype=torch.int32).item()

        _mla_mask_kernel[grid](
            q, k, v, mask, kv_cache,
            out, kv_out,
            S=S,
            R=R,
            DK=DK,
            DV=DV,
            dropout_p=dropout_p,
            seed=base_seed,
        )

        return out, kv_out


    @triton.jit
    def _mla_mask_kernel(
        Q_ptr, K_ptr, V_ptr, MASK_ptr, kv_cache_ptr,
        OUT_ptr, OUT_KV_cache_ptr,
        S: tl.constexpr,
        R: tl.constexpr,
        DK: tl.constexpr,
        DV: tl.constexpr,
        dropout_p,
        seed,
    ):
        pid = tl.program_id(axis=0)
        stride_qb = S * DK
        stride_kb = R * DK
        stride_vb = R * DV
        stride_ob = S * DV

        Q_ptr  = Q_ptr  + pid * stride_qb
        K_ptr  = K_ptr  + pid * stride_kb
        V_ptr  = V_ptr  + pid * stride_vb
        OUT_ptr = OUT_ptr + pid * stride_ob

        kv_block_ptr = tl.make_block_ptr(
            base=kv_cache_ptr + pid * DK * DV,
            shape=(DK, DV),
            strides=(DV, 1),
            offsets=(0, 0),
            block_shape=(DK, DV),
            order=(0, 1),
        )
        OUT_KV_cache_block_ptr = tl.make_block_ptr(
            base=OUT_KV_cache_ptr + pid * DK * DV,
            shape=(DK, DV),
            strides=(DV, 1),
            offsets=(0, 0),
            block_shape=(DK, DV),
            order=(0, 1),
        )
        running_kv = tl.load(kv_block_ptr)

        prev_idx = tl.full((), -1, tl.int32)

        for pos in tl.range(0, S):  # type: ignore
            tgt_idx = tl.load(MASK_ptr + pos)
            delta   = tgt_idx - prev_idx
            delta   = tl.maximum(delta, 0)
            for step in tl.range(delta):  # type: ignore
                idx  = prev_idx + 1 + step
                k_vec = tl.load(K_ptr + idx * DK + tl.arange(0, DK))
                v_vec = tl.load(V_ptr + idx * DV + tl.arange(0, DV))
                running_kv += k_vec[:, None] * v_vec[None, :]

            prev_idx = tgt_idx
            q_vec   = tl.load(Q_ptr + pos * DK + tl.arange(0, DK))

            if dropout_p > 0.0:
                dropout_scaling = 1.0 - dropout_p
                rng_seed = seed + pid * S + pos
                offs_k = tl.arange(0, DK)[:, None] * DV
                offs_v = tl.arange(0, DV)[None, :]
                offset = offs_k + offs_v
                rnd = tl.rand(rng_seed, offset)
                dropout_mask = rnd >= dropout_p
                kv_view = running_kv * dropout_mask / dropout_scaling
            else:
                kv_view = running_kv

            out_vec = tl.sum(kv_view * q_vec[:, None], axis=0)
            tl.store(OUT_ptr + pos * DV + tl.arange(0, DV), out_vec)
        tl.store(OUT_KV_cache_block_ptr, running_kv)

def multi_linear_attn(
    query,
    key,
    value,
    mask=None,
    dropout_p: float = 0.0,
    is_causal: bool = False,
    scale=None,
    enable_gqa: bool = False,
    kv_cache=None,
):
    seq_len = query.size(-2)
    B, H, dict_size, DK = key.shape
    DV = value.shape[-1]
    scale_factor = 1 / dict_size if scale is None else scale

    if mask is not None:
        assert mask.shape == (seq_len,)
        assert torch.all(0 <= mask) & torch.all(mask < dict_size)
        mask = mask.to(torch.int32)

    if is_causal:
        assert mask is None
        mask = torch.arange(dict_size-seq_len, dict_size, dtype=torch.int32)
        mask = mask.to(query.device)

    if enable_gqa:
        key = key.repeat_interleave(query.size(-3)//key.size(-3), -3)
        value = value.repeat_interleave(query.size(-3)//value.size(-3), -3)

    if kv_cache is None:
        kv_cache = torch.zeros(B, H, DK, DV, device=query.device, dtype=query.dtype)

    query = query * scale_factor
    use_triton = mask is not None and _TRITON_AVAILABLE and query.device.type == 'cuda'
    if use_triton:
        padding_needed = triton.next_power_of_2(DK) != DK or triton.next_power_of_2(DV) != DV
        if padding_needed:
            query = pad_to_pow2(query, dims=[-1])
            key = pad_to_pow2(key, dims=[-1])
            value = pad_to_pow2(value, dims=[-1])
            kv_cache = pad_to_pow2(kv_cache, dims=[-2, -1])
            out, kv_cache = _multi_linear_attn_mask_triton(
                query,
                key,
                value,
                mask,
                dropout_p,
                kv_cache,
            )
            out = out[..., :DV]
            kv_cache = kv_cache[:, :, :DK, :DV]
        else:
            out, kv_cache = _multi_linear_attn_mask_triton(
                query,
                key,
                value,
                mask,
                dropout_p,
                kv_cache,
            )
        return out, kv_cache
    elif mask is not None:
        return _multi_linear_attn_mask_torch(
            query,
            key,
            value,
            mask,
            dropout_p,
            kv_cache,
        )
    else:
        return _multi_linear_attn_no_mask(
            query,
            key,
            value,
            dropout_p,
            kv_cache
        )


import torch
from torch import Tensor
from torch.nn import Module, Parameter
from torch.nn.init import xavier_uniform_, constant_, xavier_normal_
from torch.nn.functional import _in_projection
from torch.nn.modules.linear import NonDynamicallyQuantizableLinear
from typing import Optional, Callable

from .functional import multi_linear_attn

class TrimMultiheadAttention(Module):
    """
    MultiheadAttention module using multi-linear attention instead of 
    standard scaled dot product attention.
    
    This implementation follows the same interface as PyTorch's MultiheadAttention
    but uses multi-linear attention for more efficient computation.

    Additionally implements a key-value cache for inference.
    """
    bias_k: Optional[torch.Tensor]
    bias_v: Optional[torch.Tensor]

    def __init__(
        self,
        embed_dim,
        num_heads,
        dropout=0.0,
        bias=True,
        add_bias_kv=False,
        kdim=None,
        vdim=None,
        batch_first=False,
        norm_q: Optional[Module] = None,
        norm_k: Optional[Module] = None,
        norm_v: Optional[Module] = None,
        q_weight_init: Optional[Callable[[Tensor], Tensor]] | Optional[Callable[[Tensor], None]] = None,
        k_weight_init: Optional[Callable[[Tensor], Tensor]] | Optional[Callable[[Tensor], None]] = None,
        v_weight_init: Optional[Callable[[Tensor], Tensor]] | Optional[Callable[[Tensor], None]] = None,
        scale: Optional[float] = None,
        device=None,
        dtype=None,
    ) -> None:
        if embed_dim <= 0 or num_heads <= 0:
            raise ValueError(
                f"embed_dim and num_heads must be greater than 0,"
                f" got embed_dim={embed_dim} and num_heads={num_heads} instead"
            )
        self.factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.embed_dim = embed_dim
        self.kdim = kdim if kdim is not None else embed_dim
        self.vdim = vdim if vdim is not None else embed_dim

        self.num_heads = num_heads
        self.dropout = dropout
        self.batch_first = batch_first
        self.head_dim = embed_dim // num_heads
        assert (
            self.head_dim * num_heads == self.embed_dim
        ), "embed_dim must be divisible by num_heads"

        self.q_proj_weight = Parameter(
            torch.empty((embed_dim, embed_dim), **self.factory_kwargs)
        )
        self.k_proj_weight = Parameter(
            torch.empty((embed_dim, self.kdim), **self.factory_kwargs)
        )
        self.v_proj_weight = Parameter(
            torch.empty((embed_dim, self.vdim), **self.factory_kwargs)
        )

        if bias:
            self.in_proj_bias = Parameter(torch.empty(3 * embed_dim, **self.factory_kwargs))
        else:
            self.register_parameter("in_proj_bias", None)

        self.out_proj = NonDynamicallyQuantizableLinear(
            embed_dim, embed_dim, bias=bias, **self.factory_kwargs
        )

        if add_bias_kv:
            self.bias_k = Parameter(torch.empty((1, 1, embed_dim), **self.factory_kwargs))
            self.bias_v = Parameter(torch.empty((1, 1, embed_dim), **self.factory_kwargs))
        else:
            self.bias_k = self.bias_v = None

        self.norm_q = norm_q
        self.norm_k = norm_k
        self.norm_v = norm_v

        self.q_weight_init = q_weight_init
        self.k_weight_init = k_weight_init
        self.v_weight_init = v_weight_init

        self.scale = scale
        self.kv_cache = None
        self._reset_parameters()

    def _reset_parameters(self):
        if self.q_weight_init is not None:
            self.q_weight_init(self.q_proj_weight)
        else:
            xavier_uniform_(self.q_proj_weight)
        
        if self.k_weight_init is not None:
            self.k_weight_init(self.k_proj_weight)
        else:
            xavier_uniform_(self.k_proj_weight)

        if self.v_weight_init is not None:
            self.v_weight_init(self.v_proj_weight)
        else:
            xavier_uniform_(self.v_proj_weight)

        if self.in_proj_bias is not None:
            constant_(self.in_proj_bias, 0.0)
            constant_(self.out_proj.bias, 0.0)
        if self.bias_k is not None:
            xavier_normal_(self.bias_k)
        if self.bias_v is not None:
            xavier_normal_(self.bias_v)

    def forward(
        self,
        query: Tensor,
        key: Tensor,
        value: Tensor,
        is_causal: bool = False,
        mask: Optional[Tensor] = None,
        src_key_padding_mask: Optional[Tensor] = None,
        use_kv_cache: bool = False,
        update_kv_cache: bool = False,
    ) -> Tensor:
        is_batched = query.dim() == 3

        if not is_batched:
            query, key, value = (x.unsqueeze(1) for x in (query, key, value))
        assert query.dim() == key.dim() == value.dim() == 3

        if not self.batch_first:
            query, key, value = (x.transpose(1, 0) for x in (query, key, value))

        # Get dimensions
        bsz, tgt_len, embed_dim = query.shape
        bsz, src_len, _ = key.shape

        # Input projections
        q, k, v = _in_projection(
            query, key, value,
            self.q_proj_weight, self.k_proj_weight, self.v_proj_weight,
            self.in_proj_bias[0:embed_dim] if self.in_proj_bias is not None else None,
            self.in_proj_bias[embed_dim:2*embed_dim] if self.in_proj_bias is not None else None,
            self.in_proj_bias[2*embed_dim:] if self.in_proj_bias is not None else None,
        )

        # Add bias vectors if specified
        if self.bias_k is not None and self.bias_v is not None:
            k = torch.cat([k, self.bias_k.repeat(1, bsz, 1)])
            v = torch.cat([v, self.bias_v.repeat(1, bsz, 1)])
            src_len += 1

        # Reshape for multi-head attention: (batch, seq_len, embed_dim) -> (batch, num_heads, seq_len, head_dim)
        q = q.view(bsz, tgt_len, self.num_heads, self.head_dim).transpose(1, 2)
        k = k.view(bsz, src_len, self.num_heads, self.head_dim).transpose(1, 2)
        v = v.view(bsz, src_len, self.num_heads, self.head_dim).transpose(1, 2)

        if src_key_padding_mask is not None:
            assert src_key_padding_mask.shape == (bsz, src_len)
            mask_expanded = src_key_padding_mask.unsqueeze(1).unsqueeze(3)
            mask_expanded = mask_expanded.expand(bsz, self.num_heads, src_len, self.head_dim)
            k = k.masked_fill(mask_expanded, 0.0)

        dropout_p = self.dropout if self.training else 0.0

        if use_kv_cache:
            if self.kv_cache is None:
                self.kv_cache = torch.zeros(bsz, self.num_heads, self.kdim//self.num_heads, self.vdim//self.num_heads, device=query.device)
            kv_cache = self.kv_cache
            assert kv_cache.shape == (bsz, self.num_heads, self.kdim//self.num_heads, self.vdim//self.num_heads)
        else:
            kv_cache = None

        if self.norm_q is not None:
            q = self.norm_q(q)
        if self.norm_k is not None:
            k = self.norm_k(k)
        if self.norm_v is not None:
            v = self.norm_v(v)

        attn_output, key_value_store = multi_linear_attn(
            q, k, v, 
            mask=mask,
            dropout_p=dropout_p,
            is_causal=is_causal,
            scale=self.scale,
            enable_gqa=False,
            kv_cache=kv_cache
        )

        if update_kv_cache:
            self.kv_cache = key_value_store

        # Reshape back: (batch, num_heads, tgt_len, head_dim) -> (batch, tgt_len, embed_dim)
        attn_output = attn_output.transpose(1, 2).reshape(bsz, tgt_len, embed_dim)

        # Output projection
        attn_output = torch.nn.functional.linear(
            attn_output.reshape(-1, embed_dim),
            self.out_proj.weight,
            self.out_proj.bias,
        )
        attn_output = attn_output.reshape(bsz, tgt_len, embed_dim)

        # Convert back to seq_first if needed
        if not self.batch_first:
            attn_output = attn_output.transpose(1, 0)

        # Remove batch dimension if input was unbatched
        if not is_batched:
            attn_output = attn_output.squeeze(1)

        return attn_output

    def clear_kv_cache(self):
        self.kv_cache = None


import copy
import warnings
from typing import Callable, Optional, Union

import torch
import torch.nn.functional as F
from torch import Tensor
from torch.nn.init import xavier_uniform_
from torch.nn import Module, ModuleList, Dropout, Linear, LayerNorm

from .modules import TrimMultiheadAttention


def _get_clones(module, N):
    return ModuleList([copy.deepcopy(module) for i in range(N)])


def _get_activation_fn(activation: Union[str, Callable[[Tensor], Tensor]]) -> Callable[[Tensor], Tensor]:
    if activation == "relu":
        return F.relu
    elif activation == "gelu":
        return F.gelu
    elif activation == "tanh":
        return torch.tanh
    elif activation == "sigmoid":
        return torch.sigmoid
    
    return activation


class TrimTransformerEncoderLayer(Module):
    r"""
    This is a modified version of PyTorch's TransformerEncoderLayer that uses multi-linear attention
    instead of standard scaled dot product attention for more efficient computation.

    Args:
        d_model: the number of expected features in the input (required).
        nhead: the number of heads in the multiheadattention models (required).
        dim_feedforward: the dimension of the feedforward network model (default=2048).
        dropout: the dropout value (default=0.1).
        activation: the activation function of the intermediate layer, can be a string
            ("relu" or "gelu") or a unary callable. Default: relu
        layer_norm_eps: the eps value in layer normalization components (default=1e-5).
        batch_first: If ``True``, then the input and output tensors are provided
            as (batch, seq, feature). Default: ``False`` (seq, batch, feature).
        norm_first: if ``True``, layer norm is done prior to attention and feedforward
            operations, respectively. Otherwise it's done after. Default: ``False`` (after).
        bias: If set to ``False``, ``Linear`` and ``LayerNorm`` layers will not learn an additive
            bias. Default: ``True``.
        pos_emb: the positional encoding module. Defaults to None.

    Examples::
        >>> encoder_layer = TrimTransformerEncoderLayer(d_model=512, nhead=8)
        >>> src = torch.rand(10, 32, 512)
        >>> out = encoder_layer(src)

    Alternatively, when ``batch_first`` is ``True``:
        >>> encoder_layer = TrimTransformerEncoderLayer(d_model=512, nhead=8, batch_first=True)
        >>> src = torch.rand(32, 10, 512)
        >>> out = encoder_layer(src)
    """


    def __init__(
        self,
        d_model: int,
        nhead: int,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: Union[str, Callable[[Tensor], Tensor]] = F.relu,
        layer_norm_eps: float = 1e-5,
        batch_first: bool = False,
        norm_first: bool = False,
        bias: bool = True,
        pos_emb: Optional[Module] = None,
        norm_q: Optional[Module] = None,
        norm_k: Optional[Module] = None,
        norm_v: Optional[Module] = None,
        q_weight_init: Optional[Callable[[Tensor], Tensor]] | Optional[Callable[[Tensor], None]] = None,
        k_weight_init: Optional[Callable[[Tensor], Tensor]] | Optional[Callable[[Tensor], None]] = None,
        v_weight_init: Optional[Callable[[Tensor], Tensor]] | Optional[Callable[[Tensor], None]] = None,
        scale: Optional[float] = None,
        device=None,
        dtype=None,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.self_attn = TrimMultiheadAttention(
            d_model,
            nhead,
            dropout=dropout,
            bias=bias,
            batch_first=batch_first,
            norm_q=norm_q,
            norm_k=norm_k,
            norm_v=norm_v,
            q_weight_init=q_weight_init,
            k_weight_init=k_weight_init,
            v_weight_init=v_weight_init,
            scale=scale,
            **factory_kwargs,
        )
        # Implementation of Feedforward model
        self.linear1 = Linear(d_model, dim_feedforward, bias=bias, **factory_kwargs)
        self.dropout = Dropout(dropout)
        self.linear2 = Linear(dim_feedforward, d_model, bias=bias, **factory_kwargs)
        self.batch_first = batch_first
        self.norm_first = norm_first
        self.norm1 = LayerNorm(d_model, eps=layer_norm_eps, bias=bias, **factory_kwargs)
        self.norm2 = LayerNorm(d_model, eps=layer_norm_eps, bias=bias, **factory_kwargs)
        self.dropout1 = Dropout(dropout)
        self.dropout2 = Dropout(dropout)
        self.pos_emb = pos_emb
        # Legacy string support for activation function.
        if isinstance(activation, str):
            activation = _get_activation_fn(activation)
        self.activation = activation

    def forward(
        self,
        src: Tensor,
        mask: Optional[Tensor] = None,
        src_key_padding_mask: Optional[Tensor] = None,
        is_causal: bool = False,
        use_kv_cache: bool = False,
        update_kv_cache: bool = False,
    ) -> Tensor:
        r"""Pass the input through the encoder layer.

        Args:
            src: the sequence to the encoder layer (required).
            mask: the mask tensor for multi-linear attention (optional).
                Should be of shape (seq_len,) with integer values indicating 
                the last position each token can attend to.
            src_key_padding_mask: the mask for the src keys per batch (optional).
            is_causal: If specified, applies a causal mask. Will override mask.
                Default: ``False``.
            use_kv_cache: Whether to use key-value caching for efficiency.
                Default: ``False``.
            update_kv_cache: Whether to update the key-value cache.
                Default: ``False``.

        Shape:
            - src: (N, S, E) if batch_first=True or (S, N, E) if batch_first=False
            - mask: (S,) where S is sequence length
        """
        x = src
        assert x.ndim == 3, "Input must have three axes."
        if not self.batch_first:
            x = x.permute(1, 0, 2)
        if self.pos_emb is not None:
            x = self.pos_emb(x)
        if not self.batch_first:
            x = x.permute(1, 0, 2)

        if self.norm_first:
            x = x + self._sa_block(
                self.norm1(x), mask, src_key_padding_mask, is_causal=is_causal,
                use_kv_cache=use_kv_cache, update_kv_cache=update_kv_cache
            )
            x = x + self._ff_block(self.norm2(x))
        else:
            x = self.norm1(
                x + self._sa_block(x, mask, src_key_padding_mask, is_causal=is_causal,
                                 use_kv_cache=use_kv_cache, update_kv_cache=update_kv_cache)
            )
            x = self.norm2(x + self._ff_block(x))

        return x

    # self-attention block
    def _sa_block(
        self,
        x: Tensor,
        mask: Optional[Tensor],
        src_key_padding_mask: Optional[Tensor],
        is_causal: bool = False,
        use_kv_cache: bool = False,
        update_kv_cache: bool = False,
    ) -> Tensor:
        x = self.self_attn(
            x,
            x,
            x,
            mask=mask,
            src_key_padding_mask=src_key_padding_mask,
            is_causal=is_causal,
            use_kv_cache=use_kv_cache,
            update_kv_cache=update_kv_cache,
        )
        return self.dropout1(x)

    # feed forward block
    def _ff_block(self, x: Tensor) -> Tensor:
        x = self.linear2(self.dropout(self.activation(self.linear1(x))))
        return self.dropout2(x)

    def clear_kv_cache(self):
        """Clear the key-value cache in the attention layer."""
        self.self_attn.clear_kv_cache()


class TrimTransformerEncoder(Module):
    r"""TrimTransformerEncoder is a stack of N encoder layers using multi-linear attention.

    Users can build models with multi-linear attention for improved efficiency.

    Args:
        encoder_layer: an instance of the TrimTransformerEncoderLayer() class (required).
        num_layers: the number of sub-encoder-layers in the encoder (required).
        norm: the layer normalization component (optional).

    Examples::
        >>> encoder_layer = TrimTransformerEncoderLayer(d_model=512, nhead=8)
        >>> transformer_encoder = TrimTransformerEncoder(encoder_layer, num_layers=6)
        >>> src = torch.rand(10, 32, 512)
        >>> out = transformer_encoder(src)
    """
    def __init__(
        self,
        encoder_layer: "TrimTransformerEncoderLayer",
        num_layers: int,
        norm: Optional[Module] = None,
    ) -> None:
        super().__init__()
        torch._C._log_api_usage_once(f"torch.nn.modules.{self.__class__.__name__}")
        self.layers = _get_clones(encoder_layer, num_layers)
        self.num_layers = num_layers
        self.norm = norm

    def forward(
        self,
        src: Tensor,
        mask: Optional[Tensor] = None,
        src_key_padding_mask: Optional[Tensor] = None,
        is_causal: Optional[bool] = None,
        use_kv_cache: bool = False,
        update_kv_cache: bool = False,
    ) -> Tensor:
        r"""Pass the input through the encoder layers in turn.

        Args:
            src: the sequence to the encoder (required).
            mask: the mask tensor for multi-linear attention (optional).
            src_key_padding_mask: the mask for the src keys per batch (optional).
            is_causal: If specified, applies a causal mask. Will override mask.
                Default: ``None``; try to detect a causal mask.
            use_kv_cache: Whether to use key-value caching for efficiency.
                Default: ``False``.
            update_kv_cache: Whether to update the key-value cache.
                Default: ``False``.

        Shape:
            - src: (N, S, E) if batch_first=True or (S, N, E) if batch_first=False
            - mask: (S,) where S is sequence length
        """
        output = src

        # Default is_causal to False if not specified
        if is_causal is None:
            is_causal = False

        for mod in self.layers:
            output = mod(
                output,
                mask=mask,
                is_causal=is_causal,
                src_key_padding_mask=src_key_padding_mask,
                use_kv_cache=use_kv_cache,
                update_kv_cache=update_kv_cache,
            )

        if self.norm is not None:
            output = self.norm(output)

        return output

    def clear_kv_cache(self):
        """Clear the key-value cache in all encoder layers."""
        for layer in self.layers:
            layer.clear_kv_cache()